/* Replacement FFT routine for mp3 encoding.
   Wedged in by Mike Cheng  :  http://www.cryogen.com/mikecheng
   Based upon split-radix fft by Malvar
       from book:  "Signal Processing with Lapped Transforms". Malvar, HS.
   
   The reason this fft is so damn fast, is because mp3 encoding only uses
   *real* FFTs.  The previous fft routine included in the iso source is for
   a complex->complex transform.  But if the imaginary part of the signal
   is zero (as it is for encoding sound), then you only have to do a 
   real->complex fft.  This makes this new fft routine about twice the
   speed of the old one.

   There is another real->complex fft that is about another 30% than the
   split-radix (in theory), but it may take me a lot longer to put in.

   later
   mike
*/

#include "common.h"

#define BLKSIZE 1024
#define MAXLOGM  25
#define TWOPI  6.28318530717958647692
#define SQHALF   0.707106781186547524401

// Local function prototypes
static void rsfft(float *x, int logm);
static void srfft(float *xr, float *xi, int logm);
static void srrec(float *xr, float *xi, int logm);
static void enphi(float *xrp, float *xip, float *ep, float *pp,int N);
static void enphinew(float x_real[BLKSIZE],float energy[BLKSIZE], float phi[BLKSIZE], int N);
static void BR_permute(float *x, int logm);


void fft(float x_real[BLKSIZE],float energy[BLKSIZE], float phi[BLKSIZE], int N)
{
int logm;

if (N==1024)
  logm=10;
else if (N==256)
  logm=8;
 else {
   MPegError("logm error\n");
 }
   rsfft(x_real,logm);
   enphinew(x_real,energy,phi,N);
}

static void enphinew(float x_real[BLKSIZE],float energy[BLKSIZE], float phi[BLKSIZE], int N)
{
  int i;
  float *ep, *pp, *epn, *ppn;
  float *xp1, *xpn;

	ep = energy;
	pp = phi;
	xp1 = x_real;
	xpn = &x_real[N-1];

	*ep++ = *xp1 * *xp1;
	*pp++ = atan2( 0.0, (double)(*xp1++));

	if (N==1024)
	{
		for (i=1;i<512;i++)
		{
			// Calculat energy
			*ep = *xp1 * *xp1 + *xpn * *xpn;

			// Fudge energy if below threshold
			if (*ep < 0.0005)
			{
				*ep++ = 0.0005;
				*pp++ = 0.0;
				xpn--;
				xp1++;
			}
			else 
			{
				ep++;
				#ifdef LAYER3
					if (i<8)
						*pp++ = atan2( -(double)(*xpn--), (double)(*xp1++) );
					else
					{
						pp++;
						xpn--;
						*xp1++;
					}
				#else
					*pp++ = atan2( -(double)(*xpn--), (double)(*xp1++) );
				#endif
			}
		}

		ep = &energy[513];
		pp = &phi[513];
		epn = &energy[511];
		ppn = &phi[511];

		for (i=1;i<512;i++)
		{
			*ep++ = *epn--;
			*pp++ = - *ppn--;
		}

		energy[512] = x_real[512] * x_real[512];
		phi[512] = atan2 (0.0, (double)x_real[512]);
   }
   else
   {
		for (i=1;i<N/2;i++)
		{
			*ep = *xp1 * *xp1 + *xpn * *xpn;
			if (*ep < 0.0005)
			{
				*ep++ = 0.0005;
				*pp++ = 0.0;
				xpn--;
				xp1++;
			}
			else
			{
				ep++;
				*pp++ = atan2( -(double)(*xpn--), (double)(*xp1++) );
			}
		}

		ep = &energy[N/2+1];
		pp = &phi[N/2+1];
		epn = &energy[N/2-1];
		ppn = &phi[N/2-1];
		
		for (i=1;i<N/2;i++)
		{
			*ep++ = *epn--;
			*pp++ = - *ppn--;
		}

		energy[N/2] = x_real[N/2] * x_real[N/2];
		phi[N/2] = atan2 (0.0, (double)x_real[N/2]);
	}
}

/**********************************************************
* Data unshuffling according to bit-reversed indexing.
*
* Bit reversal is done using Evans' algorithm. (Ref: DMW Evans,
* "An Improved Digit-Reversal Permutation Algorithm",
*  IEEE Trans. ASSP, Aug 1987, pp. 1120-25.
*
***********************************************************/

static int brseed[256]; /* Evans' seed table */
static int brsflg;   /* flag for table building */

static void BR_permute(float *x, int logm)
{
	register int lg2, n;
	int i,j,imax;
	int off,fj, gno, *brp;
	float tmp, *xp, *xq;

lg2 = logm >> 1;
n = 1 << lg2;

if (logm & 1) lg2++;

/* create seed table if not yet built */
if (brsflg != logm) {
  brsflg = logm;
  brseed[0] = 0;
  brseed[1] = 1;
  for(j=2; j <= lg2; j++){
    imax =  1 << (j - 1);
    for (i=0; i < imax; i++){
      brseed[i] <<= 1;
      brseed[i + imax] = brseed[i] + 1;
   }
 }
}

/* unshuffling loop */
for(off = 1; off < n; off++){
fj = n*brseed[off]; i = off; j = fj;
tmp = x[i]; x[i] = x[j]; x[j] = tmp;
xp = &x[i];
brp = &brseed[1];
for(gno = 1; gno < brseed[off]; gno++){
  xp += n;
  j = fj + *brp++;
  xq = x + j ;
  tmp = *xp; *xp = *xq; *xq = tmp;
}
}


}

/**************************************************************
*                                                             *
* Recursive part of split radix FFT algorithm
*
***************************************************************/

static void srrec(float *xr, float *xi, int logm)
{

static int m, m2, m4, m8, nel, n;
static float *xr1, *xr2, *xi1, *xi2;
static float *cn, *spcn, *smcn, *c3n, *spc3n, *smc3n;
static float tmp1, tmp2, ang, c, s;
static float *tab[MAXLOGM];

/* check range of logm */

if ((logm < 0) || (logm > MAXLOGM))
{
	sprintf(lpszError, "Error: SRFFT logm = %d is out of bounds [%d %d]\n", logm, 0, MAXLOGM);
	MPegError(lpszError);
}

/* compute trivial cases */
if (logm < 3){
if (logm == 2){
xr2 = xr + 2;
xi2= xi + 2;
tmp1 = *xr + *xr2;
*xr2 = *xr - *xr2;
*xr = tmp1;
tmp1 = *xi + *xi2;
*xi2 = *xi - *xi2;
*xi = tmp1;
xr1 = xr + 1;
xi1 = xi + 1;
xr2++;
xi2++;
tmp1 = *xr1 + *xr2;
*xr2 = *xr1 - *xr2;
*xr1 = tmp1;
tmp1 = *xi1 +  *xi2;
*xi2 = *xi1 - *xi2;
*xi1 = tmp1;
xr2 = xr + 1;
xi2 = xi + 1;
tmp1 = *xr + *xr2;
*xr2 = *xr - *xr2;
*xr = tmp1;
tmp1 = *xi + *xi2;
*xi2 = *xi - *xi2;
*xi = tmp1;
xr1 = xr +2;
xi1 = xi+2;
xr2 = xr+3;
xi2 = xi + 3;
tmp1 = *xr1 + *xi2;
tmp2 = *xi1 + *xr2;
*xi1 = *xi1 - *xr2;
*xr2 = *xr1 - *xi2;
*xr1 = tmp1;
*xi2 = tmp2;
return;
}
else if (logm == 1){
xr2 = xr + 1;
xi2 = xi + 1;
tmp1 = *xr + *xr2;
*xr2 = *xr - *xr2;
*xr = tmp1;
tmp1 = *xi + *xi2;
*xi2 = *xi - *xi2;
*xi = tmp1;
return;
}
else if (logm == 0) return;
}

/* compute a few constants */

m = 1 << logm; m2 = m / 2; m4 = m2 / 2; m8 = m4 / 2;


/* build tables of butterfly coefficients if necessary */
if ((logm >=4) && (tab[logm-4] == NULL)){
/* allocate memory for tables */
nel = m4 - 2;
if ((tab[logm-4] = (float *) calloc(6 * nel, sizeof(float)))
    == NULL){
	MPegError("");
}



/* initialize pointers */
cn = tab[logm-4]; spcn = cn + nel; smcn = spcn + nel;
c3n = smcn + nel; spc3n = c3n + nel; smc3n = spc3n + nel;


/* compute tables */
for (n = 1; n < m4; n++){
  if (n == m8) continue;
  ang = n * TWOPI / m;
  c = cos(ang); s = sin(ang);
  *cn++ = c; *spcn++ = -(s + c); *smcn++ = s - c;
  ang = 3*n*TWOPI/m;
  c = cos(ang); s = sin(ang);
  *c3n++ = c; *spc3n++ = -(s + c); *smc3n++ = s - c;
}
}


/* step 1 */
xr1 = xr; xr2 = xr1 + m2;
xi1 = xi; xi2 = xi1 + m2;

for (n =0; n < m2; n++){
  tmp1 = *xr1 + *xr2;
  *xr2 = *xr1 - *xr2;
  *xr1 = tmp1;
  tmp2 = *xi1 + *xi2;
  *xi2 = *xi1 - *xi2;
  *xi1 = tmp2;
  xr1++; xr2++; xi1++; xi2++;
} 

/* Step 2 */
xr1 = xr + m2;  xr2 = xr1 + m4;
xi1 = xi + m2;  xi2 = xi1+ m4;
for (n = 0; n < m4; n++){
  tmp1 = *xr1 + *xi2;
  tmp2 = *xi1 + *xr2;
  *xi1 = *xi1 - *xr2;
  *xr2 = *xr1 - *xi2;
  *xr1 = tmp1;
  *xi2 = tmp2;
  xr1++; xr2++; xi1++; xi2++;
}

/* Steps 3&4 */
xr1 = xr + m2; xr2 = xr1 + m4;
xi1 = xi + m2; xi2 = xi1 + m4;

if (logm >= 4) {
  nel = m4 -2;
  cn = tab[logm-4]; spcn = cn + nel; smcn = spcn + nel;
  c3n = smcn + nel; spc3n = c3n + nel; smc3n = spc3n + nel;
}

xr1++; xr2++; xi1++; xi2++;

for(n=1; n < m4; n++){
   if (n == m8){
     tmp1 = SQHALF*(*xr1 + *xi1);
     *xi1 = SQHALF*(*xi1 - *xr1);
     *xr1 = tmp1;
     tmp2 = SQHALF*(*xi2 - *xr2);
     *xi2 = -SQHALF*(*xr2 + *xi2);
     *xr2 = tmp2;
   }
   else {
     tmp2 = *cn++ *(*xr1 + *xi1);
     tmp1 = *spcn++ * *xr1 + tmp2;
     *xr1 = *smcn++ * *xi1 + tmp2;
     *xi1 = tmp1;
     tmp2 = *c3n++ * (*xr2 + *xi2);
     tmp1 = *spc3n++ * *xr2 + tmp2;
     *xr2 = *smc3n++ * *xi2 + tmp2;
     *xi2 = tmp1;
   }
xr1++; xr2++; xi1++; xi2++;
 }


/* call ssrec again with half DFT length */
srrec(xr, xi, logm -1);
/* call ssrec again twice with one quarter DFT length.
Constants have to be recomputed because they are static! */

m = 1 << logm; m2 = m/2;
srrec(xr+m2, xi+m2, logm - 2);

m = 1 << logm; m4 = 3*(m/4);
srrec(xr+m4, xi+m4, logm - 2);
}

/*****************************************************************
*  Direct transform                                               *
*
*******************************************************************/
        
static void srfft(float *xr, float *xi, int logm)
{
/* call recursive routine */
srrec(xr, xi, logm);

/* output array unshuffling using bit-reversed indices */
if (logm > 1){
  BR_permute(xr, logm);
  BR_permute(xi, logm);
}
}
/* --------------------------------------------------------------- */
static void enphi(float *xrp, float *xip, float *ep, float *pp,int N)
{
  /* put energy / phi calcs in separate routine */
int i;

 for (i=0;i<N;i++) {
    *ep = *xrp * *xrp + *xip * *xip;
    if (*ep <= 0.0005) {
        *pp++ = 0; 
        *ep++ = 0.0005;
        xrp++;
        xip++;
        }
    else { 
        *pp++ = atan2((double)(*xip++), (double)(*xrp++));
        ep++;
        }
    }
}
/* ------------------------------------------------------  */
/* spli8t radix fast fourert transform for real valued inputs.
   malvar's source code */



/* recursive part of rsfft algo. not externally callable */
static void rsrec(float *x, int logm)
{
  static int m,m2,m4,m8,nel, n;
  static float *xr1, *xr2, *xi1;
  static float *cn, *spcn, *smcn;
  static float tmp1, tmp2, ang, c, s;
  static float *tab[MAXLOGM];

  /* check rang of logm */
  if ((logm <0)|| (logm>MAXLOGM)) 
  {
	    sprintf(lpszError,"log out of range: %i\n",logm);
        MPegError(lpszError);
  }

  /* compute trivial cases */
  if (logm <2) {
    if (logm ==1) { /* length m = 2*/
      xr2 = x + 1;
      tmp1 = *x + *xr2;
      *xr2 = *x - *xr2;
      *x = tmp1;
      return;
    }
    else if (logm==0) return; /* length m =1 */
  }

  /* compute a few constants */
  m=1<<logm; m2=m/2;m4=m2/2;m8=m4/2;

  /* build tables of butterfly coefficients if necessary */
  if ((logm >= 4) && (tab[logm-4]==NULL)) {
    /* allocate memory for tables */
    nel=m4-2;
    if ((tab[logm-4] = (float *) calloc(3 * nel, sizeof(float))) == NULL ) {
      MPegError("cosine table memory error\n");
    }
    /* initialize pointers */
    cn = tab[logm-4]; spcn = cn+nel; smcn =spcn + nel;

    /* compute tables */
    for (n = 1; n<m4;n++) {
      if (n==m8) continue;
      ang = n * TWOPI/m;
      c=cos(ang); s=sin(ang);
      *cn++ = c; *spcn++ = -(s+c); *smcn++ = s-c;
    }
  }

  /* step 1 */
  xr1 = x; xr2 = xr1+m2;
  for (n=0; n< m2; n++) {
    tmp1 = *xr1 + *xr2;
    *xr2 = *xr1- *xr2;
    *xr1 = tmp1;
    xr1++; xr2++;
  }

  /* step 2 */
  xr1 = x  + m2 + m4;
  for (n=0;n<m4;n++) {
    *xr1 = - *xr1;
    xr1++;
  }

  /* steps 3 and 4 */
  xr1 = x + m2; xi1 = xr1 + m4;
  if (logm >= 4) {
    nel = m4-2;
    cn = tab[logm-4]; spcn = cn + nel; smcn = spcn + nel;
  }
  xr1++; xi1++;
  for (n = 1 ; n<m4; n++) {
    if (n == m8) {
      tmp1 = SQHALF * ( *xr1 + *xi1);
      *xi1 = SQHALF * ( *xi1 - *xr1);
      *xr1 = tmp1;
    } else {
      tmp2 = *cn++ * (*xr1 + *xi1);
      tmp1 = *spcn++ * *xr1 + tmp2;
      *xr1 = *smcn++ * *xi1 + tmp2;
      *xi1 = tmp1;
    }
    xr1++; xi1++;
  }

  /* call rsrec again with half DFT length */
  rsrec(x, logm-1);
  
  /* call complex DFT routine with quarter DFT length */
  m =1 <<logm; m2 =m/2; m4 = 3 * (m/4);
  srrec(x+m2, x+m4,logm-2);

  /* step 5. sign change and data reorder */
  m = 1 <<logm; m2 = m/2; m4=m2/2; m8=m4/2;
  xr1 = x + m2 + m4;
  xr2 = x + m -1;
  for (n=0;n<m8;n++) {
    tmp1 = *xr1;
    *xr1++ = - *xr2;
    *xr2-- = - tmp1;
  }
  xr1 = x + m2 + 1;
  xr2 = x + m - 2;
  for (n=0; n<m8;n++) {
    tmp1 = *xr1;
    *xr1++ = - *xr2;
    *xr2-- = tmp1;
    xr1++;
    xr2--;
  }
  if (logm == 2) x[3]=-x[3];
}

// Direct transform for real inputs
static void rsfft(float *x, int logm)
{
	// call recursive routine
	rsrec(x,logm);

	// output array unshuffling using bit reversed indices
	if (logm >1 )
	{
		BR_permute (x,logm);
	}
}
/*********************************************/





#include "TimeEval.h"
#include <math.h>
#include <Windows.h>
#include <stdio.h>


float TWIDDLE512R[256];
float TWIDDLE512I[256];


#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr


void CreateTwiddle512()
{
	for (int i=0;i<sizeof(TWIDDLE512R)/sizeof(TWIDDLE512R[0]);i++)
	{
		// TWIDDLE = exp(-2*i*pi/N)
		// Thus TWIDDLE^n=TWIDDLE[n]=exp(-2*pi*n/N)
		TWIDDLE512R[i]=cos(2*3.14159265355129793/512*(double)i);
		TWIDDLE512I[i]=sin(2*3.14159265355129793/512*(double)i);
	}
}

//#define TESTTWIDDLE 1
#undef TESTTWIDDLE


void four512(float* pfData)
{
	const int nL=9;
	int		n,n1,i,j,l,k,nTwoI;
	int		nLE,nLE2,nID;
		float	fIr,fIi;
	float	fBr,fBi;
	float	tempr;			// Temp for swap

	char lpszOut[255];


	static int init=0;

	if (init==0)
	{
		CreateTwiddle512();
		init++;
	}

	// n= 2^nL
	n=1<<nL;
	n1=n-1;

	// Set j to 1
	j = n >> 1;

	// No need to touch first & last
	for(i=1;i<n1;i++)
	{	
		// Don't swap twice!
		if(i<j)
		{	
			SWAP(pfData[2*i  ],pfData[2*j  ]);
			SWAP(pfData[2*i+1],pfData[2*j+1]);
		}

		// Scan counter from left looking for set bits. Clear
		// each with a subtract and propagate a carry
		// to the right. Fall out when we find a 0 bit.

		for(k=n>>1;k <= j;k >>= 1)
			j -= k;

		// Add in the propagated carry
		j += k;
	}
#define UNROLLFIRST
#ifdef UNROLLFIRST
	// Do first stage separately
	for (i=0;i<n;i+=2)
	{
		nTwoI=i*2;
		nID=nTwoI+2;

		fIr=pfData[nID  ];
		fIi=pfData[nID+1];

		// fB=fI thus to not assign, but add/subtract value directly

		// Complex multiplication and subtract 
		// of pfData[nID]=pfData[i]-pfData[nID]*fA
		pfData[nID  ]=pfData[nTwoI  ]-fIr;
		pfData[nID+1]=pfData[nTwoI+1]-fIi;

		// Complex multiplication and subtract 
		// of pfData[i]=pfData[i]+pfData[nID]*fA
		pfData[nTwoI  ]+=fIr;
		pfData[nTwoI+1]+=fIi;

	}


	for (l=1;l<nL;l++)
#else
	for (l=0;l<nL;l++)
#endif
	{
		nLE=1<<(l+1);			// nLE=2^l

		nLE2=nLE/2;

#ifdef TESTTWIDDLE
		float	fAngle;
		float	fAr,fAi;
		float	fWr,fWi;

		fAngle=-2*3.1415926/(float)nLE;
		fAr=1.0;
		fAi=0.0;

		fWr=cos(fAngle);
		fWi=sin(fAngle);
#endif

		int nTwiddleIndex=0;

#define UNROLLSECOND
#ifdef UNROLLSECOND
		j=0;
		for (i=0;i<n;i+=nLE)
		{
			// Do first stage separately
			nTwoI=2*i;
			nID=nTwoI+nLE2*2;

			fIr=pfData[nID  ];
			fIi=pfData[nID+1];

			// fB=fI thus to not assign, but add/subtract value directly

			// Complex multiplication and subtract 
			// of pfData[nID]=pfData[i]-pfData[nID]*fA
			pfData[nID  ]=pfData[nTwoI  ]-fIr;
			pfData[nID+1]=pfData[nTwoI+1]-fIi;

			// Complex multiplication and subtract 
			// of pfData[i]=pfData[i]+pfData[nID]*fA
			pfData[nTwoI  ]+=fIr;
			pfData[nTwoI+1]+=fIi;
		}

		// Adjust Twiddle Factor Index when unrolled
		nTwiddleIndex+=n/nLE;

		for (j=1;j<nLE2;j++)
#else
		for (j=0;j<nLE2;j++)
#endif
		{

			for (i=j;i<n;i+=nLE)
			{
				nTwoI=i*2;

				nID=nTwoI+nLE2*2;

				fIr=pfData[nID  ];
				fIi=pfData[nID+1];

				// Complex multiplication
				// fB=fI*fA
				fBr=fIr*TWIDDLE512R[nTwiddleIndex]-fIi*TWIDDLE512I[nTwiddleIndex];
				fBi=fIr*TWIDDLE512I[nTwiddleIndex]+fIi*TWIDDLE512R[nTwiddleIndex];

				// Complex multiplication and subtract 
				// of pfData[nID]=pfData[i]-pfData[nID]*fA
				pfData[nID  ]=pfData[nTwoI  ]-fBr;
				pfData[nID+1]=pfData[nTwoI+1]-fBi;

				// Complex multiplication and subtract 
				// of pfData[i]=pfData[i]+pfData[nID]*fA
				pfData[nTwoI  ]+=fBr;
				pfData[nTwoI+1]+=fBi;

			}
			nTwiddleIndex+=n/nLE;
		}
	}
}


void RealFour1024(float* pfData)
{
	const int n=1024;
	const int n3=n+1;

	unsigned long i,i1,i2,i3,i4,np3;
	float h1r,h1i,h2r,h2i;
	float wr,wi,wpr,wpi,wtemp,theta;

	theta=3.141592653589793F/(float) (n>>1);

	four512(pfData);
//	four1(pfData-1,512,1);
//return;
	wtemp=(float)sin(0.5*theta);
	wpr = -2.0F*wtemp*wtemp;
	wpi=(float)sin(theta);
	wr=1.0F+wpr;
	wi=wpi;


	for (i=1;i<(n>>2);i++)
	{
		// Get H[n] => RE index=2*i IM index=2*i+1
		i1=i+i;
		i2=i1+1;

		// Get H[N/2-n] => RE N-2*i IM index=N-2*i+1
		i3=n3-i2;
		i4=i3+1;

		h1r= 0.5*(pfData[i1]+pfData[i3]);
		h1i= 0.5*(pfData[i2]-pfData[i4]);

		h2r= 0.5*(pfData[i2]+pfData[i4]);
		h2i=-0.5*(pfData[i1]-pfData[i3]);

		pfData[i1]=h1r+wr*h2r-wi*h2i;
		pfData[i2]=h1i+wr*h2i+wi*h2r;

		pfData[i3]=  h1r-wr*h2r+wi*h2i;
		pfData[i4]= -h1i+wr*h2i+wi*h2r;

		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}

	pfData[0] = (h1r=pfData[0])+pfData[1];
	pfData[1] = h1r-pfData[1];
}



void MP3RealFour1024(float* pfData,float* pfEnergy,float* pfPhi)
{
	const int n=1024;
	const int n3=n+1;

	unsigned long i,i1,i2,i3,i4,np3;
	float h1r,h1i,h2r,h2i;
	float wr,wi,wpr,wpi,wtemp,theta;

	theta=3.141592653589793F/(float) (n>>1);

	four512(pfData);

	wtemp=(float)sin(0.5*theta);
	wpr = -2.0F*wtemp*wtemp;
	wpi=(float)sin(theta);
	wr=1.0F+wpr;
	wi=wpi;

	for (i=1;i<(n>>2);i++)
	{
		// Get H[n] => RE index=2*i IM index=2*i+1
		i1=i+i;
		i2=i1+1;

		// Get H[N/2-n] => RE N-2*i IM index=N-2*i+1
		i3=n3-i2;
		i4=i3+1;

		h1r= 0.5*(pfData[i1]+pfData[i3]);
		h1i= 0.5*(pfData[i2]-pfData[i4]);

		h2r= 0.5*(pfData[i2]+pfData[i4]);
		h2i=-0.5*(pfData[i1]-pfData[i3]);

		pfData[i1]=h1r+wr*h2r-wi*h2i;
		pfData[i2]=h1i+wr*h2i+wi*h2r;

		pfData[i3]=  h1r-wr*h2r+wi*h2i;
		pfData[i4]= -h1i+wr*h2i+wi*h2r;

		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	pfData[0] = (h1r=pfData[0])+pfData[1];
	pfData[1] = h1r-pfData[1];

	i=0;
	float* pfE=pfEnergy;
	float* pfP=pfPhi;
	float* pfReal=&pfData[0];
	float* pfImag=&pfData[1];

#ifndef LAYER3
	while (i<n)
	{
		*pfE=*pfReal * *pfReal+*pfImag * *pfImag;

		if (*pfE<= 0.0005F)
		{
			*pfE=0.0005F;
			*pfP++=0.0F;
		}
		else
		{
			*pfP++=atan2(-*pfImag,*pfReal);
		}

		pfE++;
		pfReal+=2;
		pfImag+=2;
		i+=2;
	}
#else
	while (i<16)
	{
		*pfE=*pfReal * *pfReal+*pfImag * *pfImag;

		if (*pfE<= 0.0005F)
		{
			*pfE=0.0005F;
			*pfP++=0.0F;
		}
		else
		{
			*pfP++=atan2(-*pfImag,*pfReal);
		}

		pfE++;
		pfReal+=2;
		pfImag+=2;
		i+=2;
	}
	while (i<n)
	{
		*pfE=*pfReal * *pfReal+*pfImag * *pfImag;

		if (*pfE<= 0.0005F)
		{
			*pfE=0.0005F;
		}

		pfE++;
		pfReal+=2;
		pfImag+=2;
		i+=2;
	}
#endif
}




void MP2RealFour1024(float* pfData,float* pfEnergy,float* pfPhi)
{
	const int n=1024;
	const int n3=n+1;

	unsigned long i,i1,i2,i3,i4,np3;
	float h1r,h1i,h2r,h2i;
	float wr,wi,wpr,wpi,wtemp,theta;

	theta=3.141592653589793F/(float) (n>>1);

	four512(pfData);

	wtemp=(float)sin(0.5*theta);
	wpr = -2.0F*wtemp*wtemp;
	wpi=(float)sin(theta);
	wr=1.0F+wpr;
	wi=wpi;

	for (i=1;i<(n>>2);i++)
	{
		// Get H[n] => RE index=2*i IM index=2*i+1
		i1=i+i;
		i2=i1+1;

		// Get H[N/2-n] => RE N-2*i IM index=N-2*i+1
		i3=n3-i2;
		i4=i3+1;

		h1r= 0.5*(pfData[i1]+pfData[i3]);
		h1i= 0.5*(pfData[i2]-pfData[i4]);

		h2r= 0.5*(pfData[i2]+pfData[i4]);
		h2i=-0.5*(pfData[i1]-pfData[i3]);

		pfData[i1]=h1r+wr*h2r-wi*h2i;
		pfData[i2]=h1i+wr*h2i+wi*h2r;

		pfData[i3]=  h1r-wr*h2r+wi*h2i;
		pfData[i4]= -h1i+wr*h2i+wi*h2r;

		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	pfData[0] = (h1r=pfData[0])+pfData[1];
	pfData[1] = h1r-pfData[1];

	i=0;
	float* pfE=pfEnergy;
	float* pfP=pfPhi;
	float* pfReal=&pfData[0];
	float* pfImag=&pfData[1];

	while (i<n)
	{
		*pfE=*pfReal * *pfReal+*pfImag * *pfImag;

		if (*pfE<= 0.0005F)
		{
			*pfE=0.0005F;
			*pfP++=0.0F;
		}
		else
		{
			*pfP++=atan2(-*pfImag,*pfReal);
		}

		pfE++;
		pfReal+=2;
		pfImag+=2;
		i+=2;
	}
}


void TestFour512()
{
	const int	nSizeFFT=512;
	float		pfData[2*nSizeFFT];
	float		pfTestData[2*nSizeFFT];
	int			i;
	int			time_ms;
	char		lpszOut[255];

	// Set test function
	for (i=0;i<nSizeFFT;i++)
	{
		pfTestData[2*i  ]=sin((double)i/(double)nSizeFFT*2.0*3.14159*5 )*100.0;
		pfTestData[2*i+1]=cos((double)i/(double)nSizeFFT*2.0*3.14159*20)*50.0;
	}

	// Reset timer
	TimeEval(0);

	for (int iTest=0;iTest<NUMTESTS;iTest++)
	{
		// Copy test data to fourier array
		memcpy(pfData,pfTestData,nSizeFFT*sizeof(float)*2);

		// Perform the transform
		four512(pfData);
	}

	// Get Time
	time_ms=TimeEval(1);

	// Print Time
	sprintf(lpszOut,"TestFour512 %d\n",time_ms);OutputDebugString(lpszOut);

	// Dump Output
	FILE* pFile=fopen("TestFour512.txt","w");

	for (i=0;i<nSizeFFT;i++)
	{
		fprintf(pFile,"%f\t%f\n",pfData[2*i],pfData[2*i+1]);
	}

	fclose(pFile);
}


void TestRealFour1024()
{
	const int	nSizeFFT=1024;
	float		pfData[nSizeFFT];
	float		pfTestData[nSizeFFT];
	int			i;
	int			time_ms;
	char		lpszOut[255];

	// Set test function
	for (i=0;i<nSizeFFT;i++)
	{
		pfTestData[i]=sin((double)i/(double)nSizeFFT*2.0*3.14159*5 )*100.0;
	}

	// Reset timer
	TimeEval(0);

	for (int iTest=0;iTest<NUMTESTS;iTest++)
	{
		// Copy test data to fourier array
		memcpy(pfData,pfTestData,nSizeFFT*sizeof(float));

		// Perform the transform
		RealFour1024(pfData);
	}

	// Get Time
	time_ms=TimeEval(1);

	// Print Time
	sprintf(lpszOut,"TestRealFour1024 %d\n",time_ms);OutputDebugString(lpszOut);

	// Dump Output
	FILE* pFile=fopen("TestRealFour1024.txt","w");

	for (i=0;i<nSizeFFT;i++)
	{
		fprintf(pFile,"%f\n",pfData[i]);
	}

	fclose(pFile);
}



void TestMP3Four()
{
	const int	nSizeFFT=1024;
	float		pfData[nSizeFFT];
	float		pfTestData[nSizeFFT];
	float		pfEnergy[nSizeFFT];
	float		pfPhi[nSizeFFT];
	int			i;
	int			time_ms;
	char		lpszOut[255];

	// Set test function
	for (i=0;i<nSizeFFT;i++)
	{
		pfTestData[i]=sin((double)i/(double)nSizeFFT*2.0*3.14159*5 )*100.0;
		pfPhi[i]=0.0F;
		pfEnergy[i]=0.0F;
	}

	// Reset timer
	TimeEval(0);

	for (int iTest=0;iTest<NUMTESTS;iTest++)
	{
		// Copy test data to fourier array
		memcpy(pfData,pfTestData,nSizeFFT*sizeof(float));

		// Perform the transform
		MP3RealFour1024(pfData,pfEnergy,pfPhi);
	}

	// Get Time
	time_ms=TimeEval(1);

	// Print Time
	sprintf(lpszOut,"TestMP3Four %d\n",time_ms);OutputDebugString(lpszOut);

	// Dump Output
	FILE* pFile=fopen("TestMP3Four.txt","w");

	for (i=0;i<nSizeFFT/2;i++)
	{
		fprintf(pFile,"%f\t%f\t%f\n",pfData[i],pfEnergy[i],pfPhi[i]);
	}

	fclose(pFile);
}







float TWIDDLE128R[256];
float TWIDDLE128I[256];



void CreateTwiddle128()
{
	for (int i=0;i<sizeof(TWIDDLE128R)/sizeof(TWIDDLE128R[0]);i++)
	{
		// TWIDDLE = exp(-2*i*pi/N)
		// Thus TWIDDLE^n=TWIDDLE[n]=exp(-2*pi*n/N)
		TWIDDLE128R[i]=cos(2*3.14159265351289793/128*(double)i);
		TWIDDLE128I[i]=sin(2*3.14159265351289793/128*(double)i);
	}
}


void four128(float* pfData)
{
	const int nL=7;
	int		n,n1,i,j,l,k,nTwoI;
	int		nLE,nLE2,nID;
		float	fIr,fIi;
	float	fBr,fBi;
	float	tempr;			// Temp for swap

	char lpszOut[255];


	static int init=0;

	if (init==0)
	{
		CreateTwiddle128();
		init++;
	}

	// n= 2^nL
	n=1<<nL;
	n1=n-1;

	// Set j to 1
	j = n >> 1;

	// No need to touch first & last
	for(i=1;i<n1;i++)
	{	
		// Don't swap twice!
		if(i<j)
		{	
			SWAP(pfData[2*i  ],pfData[2*j  ]);
			SWAP(pfData[2*i+1],pfData[2*j+1]);
		}

		// Scan counter from left looking for set bits. Clear
		// each with a subtract and propagate a carry
		// to the right. Fall out when we find a 0 bit.

		for(k=n>>1;k <= j;k >>= 1)
			j -= k;

		// Add in the propagated carry
		j += k;
	}
#define UNROLLFIRST
#ifdef UNROLLFIRST
	// Do first stage separately
	for (i=0;i<n;i+=2)
	{
		nTwoI=i*2;
		nID=nTwoI+2;

		fIr=pfData[nID  ];
		fIi=pfData[nID+1];

		// fB=fI thus to not assign, but add/subtract value directly

		// Complex multiplication and subtract 
		// of pfData[nID]=pfData[i]-pfData[nID]*fA
		pfData[nID  ]=pfData[nTwoI  ]-fIr;
		pfData[nID+1]=pfData[nTwoI+1]-fIi;

		// Complex multiplication and subtract 
		// of pfData[i]=pfData[i]+pfData[nID]*fA
		pfData[nTwoI  ]+=fIr;
		pfData[nTwoI+1]+=fIi;

	}


	for (l=1;l<nL;l++)
#else
	for (l=0;l<nL;l++)
#endif
	{
		nLE=1<<(l+1);			// nLE=2^l

		nLE2=nLE/2;

#ifdef TESTTWIDDLE
		float	fAngle;
		float	fAr,fAi;
		float	fWr,fWi;

		fAngle=-2*3.1415926/(float)nLE;
		fAr=1.0;
		fAi=0.0;

		fWr=cos(fAngle);
		fWi=sin(fAngle);
#endif

		int nTwiddleIndex=0;

#define UNROLLSECOND
#ifdef UNROLLSECOND
		j=0;
		for (i=0;i<n;i+=nLE)
		{
			// Do first stage separately
			nTwoI=2*i;
			nID=nTwoI+nLE2*2;

			fIr=pfData[nID  ];
			fIi=pfData[nID+1];

			// fB=fI thus to not assign, but add/subtract value directly

			// Complex multiplication and subtract 
			// of pfData[nID]=pfData[i]-pfData[nID]*fA
			pfData[nID  ]=pfData[nTwoI  ]-fIr;
			pfData[nID+1]=pfData[nTwoI+1]-fIi;

			// Complex multiplication and subtract 
			// of pfData[i]=pfData[i]+pfData[nID]*fA
			pfData[nTwoI  ]+=fIr;
			pfData[nTwoI+1]+=fIi;
		}

		// Adjust Twiddle Factor Index when unrolled
		nTwiddleIndex+=n/nLE;

		for (j=1;j<nLE2;j++)
#else
		for (j=0;j<nLE2;j++)
#endif
		{

			for (i=j;i<n;i+=nLE)
			{
				nTwoI=i*2;

				nID=nTwoI+nLE2*2;

				fIr=pfData[nID  ];
				fIi=pfData[nID+1];

				// Complex multiplication
				// fB=fI*fA
				fBr=fIr*TWIDDLE128R[nTwiddleIndex]-fIi*TWIDDLE128I[nTwiddleIndex];
				fBi=fIr*TWIDDLE128I[nTwiddleIndex]+fIi*TWIDDLE128R[nTwiddleIndex];

				// Complex multiplication and subtract 
				// of pfData[nID]=pfData[i]-pfData[nID]*fA
				pfData[nID  ]=pfData[nTwoI  ]-fBr;
				pfData[nID+1]=pfData[nTwoI+1]-fBi;

				// Complex multiplication and subtract 
				// of pfData[i]=pfData[i]+pfData[nID]*fA
				pfData[nTwoI  ]+=fBr;
				pfData[nTwoI+1]+=fBi;

			}
			nTwiddleIndex+=n/nLE;
		}
	}
}


void RealFour256(float* pfData)
{
	const int n=256;
	const int n3=n+1;

	unsigned long i,i1,i2,i3,i4,np3;
	float h1r,h1i,h2r,h2i;
	float wr,wi,wpr,wpi,wtemp,theta;

	theta=3.141592653589793F/(float) (n>>1);

	four128(pfData);
//	four1(pfData-1,128,1);
//return;
	wtemp=(float)sin(0.5*theta);
	wpr = -2.0F*wtemp*wtemp;
	wpi=(float)sin(theta);
	wr=1.0F+wpr;
	wi=wpi;


	for (i=1;i<(n>>2);i++)
	{
		// Get H[n] => RE index=2*i IM index=2*i+1
		i1=i+i;
		i2=i1+1;

		// Get H[N/2-n] => RE N-2*i IM index=N-2*i+1
		i3=n3-i2;
		i4=i3+1;

		h1r= 0.5*(pfData[i1]+pfData[i3]);
		h1i= 0.5*(pfData[i2]-pfData[i4]);

		h2r= 0.5*(pfData[i2]+pfData[i4]);
		h2i=-0.5*(pfData[i1]-pfData[i3]);

		pfData[i1]=h1r+wr*h2r-wi*h2i;
		pfData[i2]=h1i+wr*h2i+wi*h2r;

		pfData[i3]=  h1r-wr*h2r+wi*h2i;
		pfData[i4]= -h1i+wr*h2i+wi*h2r;

		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}

	pfData[0] = (h1r=pfData[0])+pfData[1];
	pfData[1] = h1r-pfData[1];
}



void MP3RealFour256(float* pfData,float* pfEnergy,float* pfPhi)
{
	const int n=256;
	const int n3=n+1;

	unsigned long i,i1,i2,i3,i4,np3;
	float h1r,h1i,h2r,h2i;
	float wr,wi,wpr,wpi,wtemp,theta;

	theta=3.141592653589793F/(float) (n>>1);

	four128(pfData);

	wtemp=(float)sin(0.5*theta);
	wpr = -2.0F*wtemp*wtemp;
	wpi=(float)sin(theta);
	wr=1.0F+wpr;
	wi=wpi;

	for (i=1;i<(n>>2);i++)
	{
		// Get H[n] => RE index=2*i IM index=2*i+1
		i1=i+i;
		i2=i1+1;

		// Get H[N/2-n] => RE N-2*i IM index=N-2*i+1
		i3=n3-i2;
		i4=i3+1;

		h1r= 0.5*(pfData[i1]+pfData[i3]);
		h1i= 0.5*(pfData[i2]-pfData[i4]);

		h2r= 0.5*(pfData[i2]+pfData[i4]);
		h2i=-0.5*(pfData[i1]-pfData[i3]);

		pfData[i1]=h1r+wr*h2r-wi*h2i;
		pfData[i2]=h1i+wr*h2i+wi*h2r;

		pfData[i3]=  h1r-wr*h2r+wi*h2i;
		pfData[i4]= -h1i+wr*h2i+wi*h2r;

		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	pfData[0] = (h1r=pfData[0])+pfData[1];
	pfData[1] = h1r-pfData[1];

	i=0;
	float* pfE=pfEnergy;
	float* pfP=pfPhi;
	float* pfReal=&pfData[0];
	float* pfImag=&pfData[1];

	while (i<n)
	{
		*pfE=*pfReal * *pfReal+*pfImag * *pfImag;

		if (*pfE<= 0.0005F)
		{
			*pfE=0.0005F;
			*pfP++=0.0F;
		}
		else
		{
			*pfP++=atan2(-*pfImag,*pfReal);
		}

		pfE++;
		pfReal+=2;
		pfImag+=2;
		i+=2;
	}
}




void TestFour128()
{
	const int	nSizeFFT=128;
	float		pfData[2*nSizeFFT];
	float		pfTestData[2*nSizeFFT];
	int			i;
	int			time_ms;
	char		lpszOut[255];

	// Set test function
	for (i=0;i<nSizeFFT;i++)
	{
		pfTestData[2*i  ]=sin((double)i/(double)nSizeFFT*2.0*3.14159*5 )*100.0;
		pfTestData[2*i+1]=cos((double)i/(double)nSizeFFT*2.0*3.14159*20)*50.0;
	}

	// Reset timer
	TimeEval(0);

	for (int iTest=0;iTest<NUMTESTS;iTest++)
	{
		// Copy test data to fourier array
		memcpy(pfData,pfTestData,nSizeFFT*sizeof(float)*2);

		// Perform the transform
		four128(pfData);
	}

	// Get Time
	time_ms=TimeEval(1);

	// Print Time
	sprintf(lpszOut,"TestFour128 %d\n",time_ms);OutputDebugString(lpszOut);

	// Dump Output
	FILE* pFile=fopen("TestFour128.txt","w");

	for (i=0;i<nSizeFFT;i++)
	{
		fprintf(pFile,"%f\t%f\n",pfData[2*i],pfData[2*i+1]);
	}

	fclose(pFile);
}


void TestRealFour256()
{
	const int	nSizeFFT=256;
	float		pfData[nSizeFFT];
	float		pfTestData[nSizeFFT];
	int			i;
	int			time_ms;
	char		lpszOut[255];

	// Set test function
	for (i=0;i<nSizeFFT;i++)
	{
		pfTestData[i]=sin((double)i/(double)nSizeFFT*2.0*3.14159*5 )*100.0;
	}

	// Reset timer
	TimeEval(0);

	for (int iTest=0;iTest<NUMTESTS;iTest++)
	{
		// Copy test data to fourier array
		memcpy(pfData,pfTestData,nSizeFFT*sizeof(float));

		// Perform the transform
		RealFour256(pfData);
	}

	// Get Time
	time_ms=TimeEval(1);

	// Print Time
	sprintf(lpszOut,"TestRealFour256 %d\n",time_ms);OutputDebugString(lpszOut);

	// Dump Output
	FILE* pFile=fopen("TestRealFour256.txt","w");

	for (i=0;i<nSizeFFT;i++)
	{
		fprintf(pFile,"%f\n",pfData[i]);
	}

	fclose(pFile);
}


void TestMP3FourS()
{
	const int	nSizeFFT=256;
	float		pfData[nSizeFFT];
	float		pfTestData[nSizeFFT];
	float		pfEnergy[nSizeFFT];
	float		pfPhi[nSizeFFT];
	int			i;
	int			time_ms;
	char		lpszOut[255];

	// Set test function
	for (i=0;i<nSizeFFT;i++)
	{
		pfTestData[i]=sin((double)i/(double)nSizeFFT*2.0*3.14159*5 )*100.0;
		pfPhi[i]=0.0F;
		pfEnergy[i]=0.0F;
	}

	// Reset timer
	TimeEval(0);

	for (int iTest=0;iTest<NUMTESTS;iTest++)
	{
		// Copy test data to fourier array
		memcpy(pfData,pfTestData,nSizeFFT*sizeof(float));

		// Perform the transform
		MP3RealFour256(pfData,pfEnergy,pfPhi);
	}

	// Get Time
	time_ms=TimeEval(1);

	// Print Time
	sprintf(lpszOut,"TestMP3FourS %d\n",time_ms);OutputDebugString(lpszOut);

	// Dump Output
	FILE* pFile=fopen("TestMP3FourS.txt","w");

	for (i=0;i<nSizeFFT/2;i++)
	{
		fprintf(pFile,"%f\t%f\t%f\n",pfData[i],pfEnergy[i],pfPhi[i]);
	}

	fclose(pFile);
}

